#include <assert.h>
#include <locale.h>
#include <stddef.h>
#include <ctype.h>
#include <math.h>
#include <stdio.h>
#include <errno.h>
#include <setjmp.h>
#include <stdlib.h>
#include <float.h>
#include <signal.h>
#include <string.h>
#include <limits.h>
#include <stdarg.h>
#include <time.h>

int i;
